import asyncio
import time
from concurrent.futures import ProcessPoolExecutor
from typing import Type, Iterable, Callable, Coroutine
import logging

from .executor import Executor
from ..core.config import Config
from ..core.meta_trader import MetaTrader
from .symbol import Symbol as Symbol
from .strategy import Strategy as Strategy

logger = logging.getLogger(__name__)


class Bot:
    """The bot class. Creates a bot instance to run strategies.

    Attributes:
        executor: A thread executor.
        config (Config): Config instance
        mt (MetaTrader): MetaTrader instance
    """

    config: Config
    executor: Executor
    mt: MetaTrader
    strategies: list[Strategy]

    def __init__(self):
        self.config = Config(bot=self)
        self.executor = Executor()
        self.mt5 = MetaTrader()
        self.strategies = []

    @classmethod
    def process_pool(cls, bots: dict[Callable:dict] = None, num_workers: int = None):
        """Run multiple bots in parallel using a ProcessPoolExecutor. Each bot should be a callable that accepts
        keyword arguments only.

        Args:
            bots (dict): A dictionary of bots to run with their respective keyword arguments as a dictionary
            num_workers (int): Number of workers to run the bots
        """
        num_workers = num_workers or len(bots) + 1
        with ProcessPoolExecutor(max_workers=num_workers) as executor:
            for bot, kwargs in bots.items():
                executor.submit(bot, **kwargs)

    async def initialize(self):
        """Prepares the bot by signing in to the trading account and initializing the symbols for each strategy.
        Only strategies with successfully initialized symbols will be added to the executor. Starts the global task queue.

        Raises:
            SystemExit if sign in was not successful
        """
        try:
            await self.mt5.initialize()
            login = await self.mt5.login()
            if not login:
                logger.critical("Unable to sign in to MetaTrder 5 Terminal")
                raise Exception("Unable to sign in to MetaTrader 5 Terminal")
            logger.info("Login Successful")
            await self.init_strategies()
            self.add_coroutine(coroutine=self.config.task_queue.run, on_separate_thread=True)
            self.add_coroutine(coroutine=self.executor.exit)

            if len(self.executor.strategy_runners) == 0:
                logger.warning("No strategies were added to the bot. Exiting in one second")
                await asyncio.sleep(1)
                self.config.shutdown = True
        except Exception as err:
            logger.error("%s: Bot initialization failed", err)
            raise SystemExit

    def initialize_sync(self):
        """Prepares the bot by signing in to the trading account and initializing the symbols for each strategy.
        Only strategies with successfully initialized symbols will be added to the executor.
        Starts the global task queue.

        Raises:
            SystemExit if sign in was not successful
        """
        try:
            self.mt5.initialize_sync()
            login = self.mt5.login_sync()
            if not login:
                logger.critical("Unable to sign in to MetaTrder 5 Terminal")
                raise Exception("Unable to sign in to MetaTrader 5 Terminal")
            logger.info("Login Successful")
            self.init_strategies_sync()
            self.add_coroutine(coroutine=self.config.task_queue.run, on_separate_thread=True)
            self.add_coroutine(coroutine=self.executor.exit)

            if len(self.executor.strategy_runners) == 0:
                logger.warning("No strategies were added to the bot. Exiting in one second")
                time.sleep(1)
                self.config.shutdown = True
        except Exception as err:
            logger.error("%s: Bot initialization failed", err)
            raise SystemExit

    def add_function(self, *, function: Callable[..., ...], **kwargs):
        """Add a function to the executor.

        Args:
            function (Callable): A function to be executed
        """
        self.executor.add_function(function=function, kwargs=kwargs)

    def add_coroutine(self, *, coroutine: Callable[..., ...] | Coroutine, on_separate_thread=False, **kwargs):
        """Add a coroutine to the executor.

        Args:
            coroutine (Coroutine): A coroutine to be executed
            on_separate_thread (bool): Run the coroutine
        Returns:

        """
        self.executor.add_coroutine(coroutine=coroutine, kwargs=kwargs, on_separate_thread=on_separate_thread)

    def execute(self):
        """Start the bot in sync mode"""
        self.initialize_sync()
        if self.config.shutdown is False:
            self.executor.execute()

    async def start(self):
        """Initialize the bot and call the executor it."""
        await self.initialize()
        if self.config.shutdown is False:
            self.executor.execute()

    def add_strategy(self, *, strategy: Strategy):
        """Add a strategy to the list of strategies.

        Args:
            strategy (Strategy): A Strategy instance to run on bot

        Notes:
            Make sure the symbol has been added to the market
        """
        self.strategies.append(strategy)

    def add_strategies(self, *, strategies: Iterable[Strategy]):
        """Add multiple strategies at the same time

        Args:
            strategies: A list of strategies
        """
        [self.add_strategy(strategy=strategy) for strategy in strategies]

    def add_strategy_all(
        self, *, strategy: Type[Strategy], params: dict | None = None, symbols: list[Symbol] = None, **kwargs
    ):
        """Use this to run a single strategy on multiple symbols with the same parameters and keyword arguments.

        Keyword Args:
            strategy (Strategy): Strategy class
            params (dict): A dictionary of parameters for the strategy
            symbols (list): A list of symbols to run the strategy on
        """
        [self.add_strategy(strategy=strategy(symbol=symbol, params=params, **kwargs)) for symbol in symbols]

    async def init_strategy(self, *, strategy: Strategy) -> bool:
        """Initialize a single strategy. This method is called internally by the bot."""
        res = await strategy.symbol.initialize()
        if res:
            self.executor.add_strategy(strategy=strategy)
        return res

    async def init_strategies(self):
        """Initialize the symbols for the current trading session. This method is called internally by the bot."""
        tasks = [self.init_strategy(strategy=strategy) for strategy in self.strategies]
        await asyncio.gather(*tasks)

    def init_strategy_sync(self, *, strategy: Strategy) -> bool:
        """Initialize a single strategy. This method is called internally by the bot."""
        res = strategy.symbol.initialize_sync()
        if res:
            self.executor.add_strategy(strategy=strategy)
        return res

    def init_strategies_sync(self):
        """Initialize the symbols for the current trading session. This method is called internally by the bot."""
        [self.init_strategy_sync(strategy=strategy) for strategy in self.strategies]



from dataclasses import dataclass, field, fields
import pickle
from pathlib import Path
from datetime import datetime, UTC
from logging import getLogger
from typing import NamedTuple, Iterable

import MetaTrader5
from numpy import ndarray

from ..meta_trader import MetaTrader
from ..config import Config
from ..constants import TimeFrame
from ..task_queue import TaskQueue, QueueItem
from ..._utils import backoff_decorator

logger = getLogger(__name__)


class Cursor(NamedTuple):
    """A cursor to iterate over the data. Marks the current position."""
    index: int
    time: int


@dataclass
class BackTestData:
    """The data class to store the backtesting data.

    Attributes:
        name (str): The name of the backtest data.
        terminal (dict): The terminal information.
        version (tuple): The version of the terminal.
        account (dict): The account information.
        symbols (dict): The symbols information.
        ticks (dict): The ticks data.
        rates (dict): The rates data.
        span (range): The range of the data.
        range (range): The range of the data.
        orders (dict): The orders data.
        deals (dict): The deals data.
        positions (dict): The positions data.
        open_positions (set): The open positions.
        cursor (Cursor): The cursor to iterate over the data.
        margins (dict): The margins data.
        fully_loaded (bool): A flag to indicate if the data is fully loaded
    """

    name: str = ""
    terminal: dict[str, [str | int | bool | float]] = field(default_factory=dict)
    version: tuple[int, int, str] = (0, 0, "")
    account: dict = field(default_factory=dict)
    symbols: dict[str, dict] = field(default_factory=dict)
    ticks: dict[str, ndarray] = field(default_factory=dict)
    rates: dict[str, dict[int, ndarray]] = field(default_factory=dict)
    span: range = range(0)
    range: range = range(0)
    orders: dict[int, dict] = field(default_factory=lambda: {})
    deals: dict[int, dict] = field(default_factory=lambda: {})
    positions: dict[int, dict] = field(default_factory=lambda: {})
    open_positions: set[int, ...] = field(default_factory=lambda: set())
    cursor: Cursor = None
    margins: dict[int, float] = field(default_factory=lambda: {})
    fully_loaded: bool = True

    def __str__(self):
        return f"{self.name}"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"

    def set_attrs(self, **kwargs):
        """Set the attributes of the class on the instance."""
        [setattr(self, k, v) for k, v in kwargs.items() if k in self.fields]

    @property
    def fields(self):
        """A list of the fields of the class."""
        return [f.name for f in fields(self)]


class GetData:
    """A class to get the backtesting data from the MetaTrader5 terminal.

    Attributes:
        start (datetime): The start date of the data.
        end (datetime): The end date of the data.
        symbols (Sequence[str]): The symbols to get the data for.
        timeframes (Sequence[TimeFrame]): The timeframes to get the data for.
        name (str): The name of the backtest data.
        range (range): The range of the data.
        span (range): The span of the data.
        data (BackTestData): The backtesting data.
        mt5 (MetaTrader): The MetaTrader5 instance.
        task_queue (TaskQueue): The task queue to handle the requests.
    """

    data: BackTestData

    def __init__(
        self, *, start: datetime, end: datetime, symbols: Iterable[str], timeframes: Iterable[TimeFrame], name: str = ""
    ):
        """
        Get the backtesting data from the MetaTrader5 terminal.

        Args:
            start (datetime): The start date of the data.
            end (datetime): The end date of the data.
            symbols (Sequence[str]): The symbols to get the data for.
            timeframes (Sequence[TimeFrame]): The timeframes to get the data for.
            name (str): The name of the backtest data.
        """
        self.config = Config()
        self.start = start.astimezone(tz=UTC)
        self.end = end.astimezone(tz=UTC)
        self.symbols = set(symbols)
        self.timeframes = set(timeframes)
        self.name = name or f"{start:%d-%m-%y}_{end:%d-%m-%y}"
        span_start = int(self.start.timestamp())
        span_end = int(self.end.timestamp())
        self.range = range(0, span_end - span_start)
        self.span = range(span_start, span_end)
        self.data = BackTestData(name=self.name, span=self.span, range=self.range)
        self.mt5 = MetaTrader()
        self.task_queue = TaskQueue(workers=500, mode="finite", on_exit="cancel")

    @classmethod
    def pickle_data(cls, *, data: BackTestData, name: str | Path):
        """Pickle the data to a file.

        Args:
            data (BackTestData): The data to pickle.
            name (str | Path): The name of the file to pickle the data to.
        """
        try:
            with open(name, "wb") as fo:
                pickle.dump(data, fo, protocol=pickle.HIGHEST_PROTOCOL)
        except Exception as err:
            logger.error(f"Error in dump_data: {err}")

    @classmethod
    def load_data(cls, *, name: str | Path) -> BackTestData:
        """Load the data from a file.

        Args:
            name (str | Path): The name of the file to load the data from.
        """
        try:
            with open(name, "rb") as fo:
                data = pickle.load(fo)
                return data
        except Exception as err:
            logger.error(f"Error: {err}")

    def save_data(self, *, name: str | Path = ""):
        """Save the data to a file.

        Args:
            name (str | Path): The name of the file to save the data to. If not provided, the name of the data is used.
        """
        name = name or (self.name + ".pkl" if not self.name.endswith(".pkl") else self.name)
        name = Path(self.config.backtest_dir) / name if not isinstance(name, Path) else name
        with open(name, "wb") as fo:
            pickle.dump(self.data, fo, protocol=pickle.HIGHEST_PROTOCOL)

    async def get_data(self, workers: int = None):
        """Use the task queue to get the data from the MetaTrader5 terminal.

        Args:
            workers (int): The number of workers to use in the task queue. If not provided, the default number of workers
             is used.
        """
        if workers:
            self.task_queue.workers = workers

        q_items = [
            QueueItem(self.get_symbols_rates),
            QueueItem(self.get_symbols_ticks),
            QueueItem(self.get_symbols_info),
        ]

        [self.task_queue.add(item=item, priority=0, must_complete=True) for item in q_items]

        if not self.data.account:
            self.task_queue.add(item=QueueItem(self.get_account_info), must_complete=True)

        if not self.data.terminal:
            self.task_queue.add(item=QueueItem(self.get_terminal_info), must_complete=True)

        if not self.data.version:
            self.task_queue.add(item=QueueItem(self.get_version), must_complete=True)

        await self.task_queue.run()

        if self.data.fully_loaded is False:
            logger.warning("Data not fully loaded")
            self.data = BackTestData(name=self.name, span=self.span, range=self.range, fully_loaded=False)

    async def get_terminal_info(self):
        terminal = await self.mt5.terminal_info()
        if terminal is None:
            self.data.fully_loaded = False
            self.task_queue.stop = True
        terminal = terminal._asdict()
        self.data.set_attrs(terminal=terminal)

    async def get_version(self):
        version = await self.mt5.version()
        if version is None:
            self.data.fully_loaded = False
            self.task_queue.stop = True
        self.data.set_attrs(version=version)

    @backoff_decorator
    async def get_account_info(self):
        res = await self.mt5.account_info()
        if res is None:
            self.data.fully_loaded = False
            self.task_queue.stop = True
        res = res._asdict()
        self.data.set_attrs(account=res)

    async def get_symbols_info(self):
        [
            self.task_queue.add(item=QueueItem(self.get_symbol_info, symbol=symbol))
            for symbol in self.symbols
            if self.data.symbols.get(symbol) is None
        ]

    async def get_symbols_ticks(self):
        [
            self.task_queue.add(item=QueueItem(self.get_symbol_ticks, symbol=symbol))
            for symbol in self.symbols
            if self.data.ticks.get(symbol) is None
        ]

    async def get_symbols_rates(self):
        [
            self.task_queue.add(item=QueueItem(self.get_symbol_rates, symbol=symbol, timeframe=timeframe), priority=4)
            for symbol in self.symbols
            for timeframe in self.timeframes
            if self.data.rates.get(symbol, {}).get(timeframe) is None
        ]

    @backoff_decorator
    async def get_symbol_info(self, *, symbol: str):
        res = await self.mt5.symbol_info(symbol)
        if res is None:
            self.data.fully_loaded = False
            self.task_queue.stop = True
        self.data.symbols[symbol] = res._asdict()

    @backoff_decorator
    async def get_symbol_ticks(self, *, symbol: str):
        res = await self.mt5.copy_ticks_range(symbol, self.start, self.end, MetaTrader5.COPY_TICKS_ALL)
        if res is None:
            self.data.fully_loaded = False
            self.task_queue.stop = True
        self.data.ticks[symbol] = res

    @backoff_decorator
    async def get_symbol_rates(self, *, symbol: str, timeframe: TimeFrame):
        res = await self.mt5.copy_rates_range(symbol, timeframe, self.start, self.end)
        if res is None:
            self.data.fully_loaded = False
            self.task_queue.stop = True
        self.data.rates.setdefault(symbol, {})[int(timeframe)] = res



"""The base class for creating strategies."""

import asyncio
from time import time
from typing import TypeVar
from abc import ABC, abstractmethod
from datetime import time as dtime
from logging import getLogger

from ..core.meta_trader import MetaTrader
from ..core import Config
from ..core.exceptions import StopTrading
from ..core.meta_backtester import MetaBackTester
from ..core.backtesting.backtest_controller import BackTestController
from .sessions import Sessions, Session
from .symbol import Symbol as _Symbol

Symbol = TypeVar("Symbol", bound=_Symbol)
logger = getLogger(__name__)


class Strategy(ABC):
    """The base class for creating strategies.

    Attributes:
        name (str): The name of the strategy.
        symbol (Symbol): The Financial Instrument as a Symbol Object
        parameters (Dict): A dictionary of parameters for the strategy.
        sessions (Sessions): The sessions to use for the strategy.
        running (bool): A flag to indicate if the strategy is running.
        backtest_controller (BackTestController): A controller for running the backtester.
        current_session (Session): The current session.
        mt5 (MetaTrader|MetaBackTester): The MetaTrader object.
        config (Config): The config object.

    Notes:
        Define the name of a strategy as a class attribute. If not provided, the class name will be used as the name.
    """

    name: str
    symbol: Symbol
    sessions: Sessions
    mt5: MetaTrader | MetaBackTester
    config: Config
    running: bool
    parameters = {}
    backtest_controller = BackTestController
    current_session = Session

    def __init__(self, *, symbol: Symbol, params: dict = None, sessions: Sessions = None, name=""):
        """Initiate the parameters dict and add name and symbol fields.
        Use class name as strategy name if name is not provided

        Args:
            symbol (Symbol): The Financial instrument
            params (Dict): Trading strategy parameters
        """
        self.parameters = {**self.parameters} | (params or {})
        self.symbol = symbol
        self.name = name or self.__class__.__name__
        self.parameters["symbol"] = symbol.name
        self.parameters["name"] = self.name
        self.running = True
        self.sessions = sessions or Sessions(sessions=[Session(start=0, end=dtime(hour=23, minute=59, second=59))])
        self.config = Config()
        self.mt5 = MetaTrader() if self.config.mode != "backtest" else MetaBackTester()
        self.backtest_controller = BackTestController()

    def __repr__(self):
        return f"{self.name}({self.symbol!r})"

    def __getattr__(self, item):
        if item in self.parameters:
            return self.parameters[item]
        raise AttributeError(f"{item} not an attribute of {self.name}")

    def __setattr__(self, key, value):
        if key in self.parameters:
            self.parameters[key] = value
        super().__setattr__(key, value)

    async def __aenter__(self):
        await self.sessions.check()
        self.running = True
        self.current_session = self.sessions.current_session

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        try:
            await self.current_session.close() if self.current_session else ...
            self.running = False
        except Exception as err:
            logger.error(f"Error: {err}")

    async def initialize(self):
        """Perform any initialization tasks here."""
        ...

    @staticmethod
    async def live_sleep(*, secs: float):
        """Sleep for the needed amount of seconds in between requests to the terminal.
        computes the accurate amount of time needed to sleep ensuring that the next request is made at the start of
        a new bar and making cooperative multitasking possible.

        Args:
            secs (float): The time in seconds. Usually the timeframe you are trading on.
        """
        mod = time() % secs
        secs = secs - mod if mod != 0 else mod
        await asyncio.sleep(secs + 0.1)

    async def sleep(self, *, secs: float):
        """Sleep for the needed amount of seconds in between requests to the terminal.
        computes the accurate amount of time needed to sleep ensuring that the next request is made at the start of
        a new bar and making cooperative multitasking possible.

        Args:
            secs (float): The time in seconds. Usually the timeframe you are trading on.
        """
        if self.config.mode == "backtest":
            await self.backtest_sleep(secs=secs)
        else:
            await self.live_sleep(secs=secs)

    async def delay(self, *, secs: float):
        """Sleep for the input amount of seconds"""
        if self.config.mode == "backtest":
            await self._backtest_sleep(secs=secs)
        else:
            await asyncio.sleep(secs)

    async def _backtest_sleep(self, *, secs: float):
        try:
            # if self.backtest_controller.parties == 2:
                # this is not needed, as the backtest_engine iterator will handle this
                # steps = int(secs) // self.config.backtest_engine.speed
                # steps = max(steps, 1)
                # self.config.backtest_engine.fast_forward(steps=steps)
                # self.backtest_controller.wait()

            if self.backtest_controller.parties >= 2:
                _time = self.config.backtest_engine.cursor.time + secs
                while _time > self.config.backtest_engine.cursor.time:
                    self.backtest_controller.wait()
            else:
                self.backtest_controller.wait()
        except Exception as err:
            self.backtest_controller.wait()
            logger.error("Error: %s in backtest_sleep", err)

    async def backtest_sleep(self, *, secs: float):
        """Sleep for the needed amount of seconds in between requests to the terminal.

        Args:
            secs (float): The time in seconds. Usually the timeframe you are trading on.
        """
        try:
            _time = self.config.backtest_engine.cursor.time
            mod = _time % secs
            secs = secs - mod if mod != 0 else mod
            await self._backtest_sleep(secs=secs)
        except Exception as err:
            logger.error("Error: %s in backtest_sleep", err)

    async def run_strategy(self):
        """Run the strategy."""
        if self.config.mode == "live":
            await self.live_strategy()
        elif self.config.mode == "backtest":
            await self.backtest_strategy()

    async def live_strategy(self):
        """Run the strategy."""
        async with self as _:
            logger.info("Running %s strategy on %s", self.name, self.symbol.name)
            await self.initialize()
            while self.running:
                try:
                    await self.sessions.check()
                    await self.trade()

                except StopTrading:
                    self.running = False
                    break

                except asyncio.CancelledError:
                    self.running = False
                    break

                except Exception as err:
                    logger.error("Error: %s in live_strategy", err)
                    self.running = False
                    break

    async def backtest_strategy(self):
        """Backtest the strategy."""
        async with self as _:
            logger.info("Testing %s strategy on %s with Backtester", self.name, self.symbol.name)
            await self.initialize()
            while self.running:
                try:
                    await self.sessions.check()
                    self.backtest_controller.wait()
                    await self.test()
                except StopTrading:
                    self.running = False
                    break
                except Exception as err:
                    logger.error(f"Error: {err} in backtest_strategy")
                    return

    @abstractmethod
    async def trade(self):
        """Place trades using this method. This is the main method of the strategy.
        It will be called by the strategy runner.
        """
        raise NotImplementedError("Implement this method in your subclass")

    async def test(self):
        await self.trade()



import asyncio
import time
import random
from typing import Coroutine, Callable, Literal
from logging import getLogger
from signal import SIGINT, signal

logger = getLogger(__name__)


class QueueItem:
    """A class to represent a task item in the queue.

        Attributes:
            - `task_item` (Callable | Coroutine): The task to run.

            - `args` (tuple): The arguments to pass to the task

            - `kwargs` (dict): The keyword arguments to pass to the task

            - `must_complete` (bool): A flag to indicate if the task must complete before the queue stops. Default is False.

            - `time` (int): The time the task was added to the queue.
    """
    def __init__(self, task_item: Callable | Coroutine, *args, **kwargs):
        self.task_item = task_item
        self.args = args
        self.kwargs = kwargs
        self.must_complete = False
        self.time = time.time_ns()

    def __hash__(self):
        return self.time

    def __lt__(self, other):
        return self.time < other.time

    def __eq__(self, other):
        return self.time == other.time

    def __le__(self, other):
        return self.time <= other.time

    async def run(self):
        try:
            if asyncio.iscoroutinefunction(self.task_item):
                return await self.task_item(*self.args, **self.kwargs)
            else:
                return await asyncio.to_thread(self.task_item, *self.args, **self.kwargs)
        except asyncio.CancelledError:
            logger.debug("Task %s with args %s and %s was cancelled",
                           self.task_item.__name__, self.args, self.kwargs)
        except Exception as err:
            logger.error("Error %s occurred in %s with args %s and %s",
                         err, self.task_item.__name__, self.args, self.kwargs)


class TaskQueue:
    queue_task: asyncio.Task
    start_time: float

    def __init__(self, *, size: int = 0, workers: int = 10, queue: asyncio.Queue = None, queue_timeout: int = 0,
                 on_exit: Literal['cancel', 'complete_priority'] = 'complete_priority', absolute_timeout: int = 0,
                 mode: Literal['finite', 'infinite'] = 'finite', worker_timeout: int = 1):
        self.queue = queue or asyncio.PriorityQueue(maxsize=size)
        self.workers = workers
        self.worker_tasks: dict[int|float, asyncio.Task] = {}
        self.queue_timeout = queue_timeout
        self.absolute_timeout = absolute_timeout
        self.stop = False
        self.on_exit = on_exit
        self.mode = mode
        self.worker_timeout = worker_timeout
        self.queue_task_cancelled = False
        self.start_time = time.perf_counter()
        signal(SIGINT, self.sigint_handle)

    def add(self, *, item: QueueItem, priority=3, must_complete=False):
        """Add a task to the queue.

        Args:
            item (QueueItem): The task to add to the queue.
            priority (int): The priority of the task. Default is 3.
            must_complete (bool): A flag to indicate if the task must complete before the queue stops. Default is False.
        """
        try:
            if self.stop:
                return
            item.must_complete = must_complete
            if isinstance(self.queue, asyncio.PriorityQueue):
                item = (priority, item)
            self.queue.put_nowait(item)
        except asyncio.QueueFull:
            logger.error("Queue is full")

    async def worker(self, wid: int = None):
        """Worker function to run tasks in the queue."""
        while True:
            try:
                if self.queue_task_cancelled or not self.check_timeout():
                    self.remove_worker(wid)
                    break

                if self.mode == 'infinite' and self.queue.qsize() <= 1:
                    dummy = QueueItem(self.dummy_task)
                    self.add(item=dummy)

                if isinstance(self.queue, asyncio.PriorityQueue):
                    _, item = self.queue.get_nowait()

                else:
                    item = self.queue.get_nowait()

                if self.stop is False or item.must_complete:
                    await item.run()

                self.queue.task_done()

                if self.stop and (self.on_exit == 'cancel' or len(self.worker_tasks) <= 1):
                    self.cancel()

                await self.add_workers()

            except asyncio.QueueEmpty:
                if self.stop:
                    self.remove_worker(wid)
                    break

                if self.mode == 'finite':
                    self.remove_worker(wid)
                    break

            except asyncio.CancelledError:
                break

            except Exception as err:
                logger.error("%s: Error occurred in worker", err)
                self.remove_worker(wid)
                break

    def start_timer(self, *, queue_timeout: int = None, absolute_timeout: int = None, start=False):
        self.queue_timeout = queue_timeout or self.queue_timeout
        self.absolute_timeout = absolute_timeout or self.absolute_timeout
        if start:
            self.start_time = time.perf_counter()

    def check_timeout(self):
        if self.queue_timeout and (time.perf_counter() - self.start_time) > self.queue_timeout:
            if self.on_exit == 'cancel':
                self.stop = True
                self.cancel()
                return False
            else:
                self.stop = True
                self.queue_timeout = 0
                return True
        if self.absolute_timeout and (time.perf_counter() - self.start_time) > self.absolute_timeout:
            self.stop = True
            self.cancel()
            return False
        return True

    async def dummy_task(self):
        await asyncio.sleep(self.worker_timeout)

    def remove_worker(self, wid: int):
        try:
            task = self.worker_tasks.pop(wid, None)
            if task is not None:
                task.cancel()
        except Exception as err:
            logger.debug("%s: Error occurred in removing worker %d", err, wid)
        except asyncio.CancelledError as _:
            ...

    async def add_workers(self, no_of_workers: int = None):
        """Create workers for running queue tasks."""
        if no_of_workers is None:
            queue_size = self.queue.qsize()
            req_workers = queue_size - len(self.worker_tasks)
            if req_workers > 1:
                no_of_workers = req_workers
            else:
                return

        ri = lambda : random.randint(999, 999_999_999) # random id
        ct = lambda ti: asyncio.create_task(self.worker(wid=ti), name=ti) # create task
        wr = range(no_of_workers)
        [self.worker_tasks.setdefault(wi:=ri(), ct(wi)) for _ in wr]

    async def watch(self):
        while True:
            await asyncio.sleep(1)
            if (time.perf_counter() - self.start_time) > self.absolute_timeout:
                self.stop = True
                self.cancel()
                break

    async def run(self, queue_timeout: int = None, absolute_timeout: int = None):
        """Run the queue until all tasks are completed or the timeout is reached.

        Args:
            queue_timeout (int): The maximum time to wait for the queue to complete. Default is 0.
            absolute_timeout (int): The maximum time to run the queue.
            This timeout overrides the timeout attribute of the queue instance.
            The queue stops when the timeout is reached, and the remaining tasks are handled based on the
            `on_exit` attribute. If the timeout is 0, the queue will run until all tasks are completed or the queue
            is stopped.
        """
        try:
            self.start_timer(queue_timeout=queue_timeout, absolute_timeout=absolute_timeout, start=True)
            await self.add_workers(no_of_workers=self.workers)
            self.queue_task = asyncio.create_task(self.queue.join())
            if self.absolute_timeout:
                asyncio.create_task(self.watch())
            await self.queue_task

        except asyncio.TimeoutError:
            logger.warning("Timeout occurred after %d seconds, %d tasks remaining",
                           time.perf_counter() - self.start_time, self.queue.qsize())

        except asyncio.CancelledError:
            logger.warning("Task Queue Cancelled after %d seconds, %d tasks remaining",
                           time.perf_counter() - self.start_time, self.queue.qsize())
        except Exception as err:
            logger.warning("%s occurred after %d seconds, %d tasks remaining",
                           err, time.perf_counter() - self.start_time, self.queue.qsize())
        finally:
            logger.info("Tasks completed after %d seconds, %d tasks remaining",
                           time.perf_counter() - self.start_time, self.queue.qsize())

    def cancel_all_workers(self):
        try:
            wids = list(self.worker_tasks.keys())
            [self.remove_worker(wid) for wid in wids]
        except Exception as err:
            logger.error("%s: Error occurred in cancelling workers", err)

    def cancel(self):
        try:
            self.queue_task.cancel()
            self.queue_task_cancelled = True
            self.cancel_all_workers()
        except asyncio.CancelledError as _:
            ...
        except Exception as err:
            logger.error("%s: Error occurred in cancelling queue", err)

    def sigint_handle(self, sig, frame):
        if self.stop is False:
            self.stop = True
        else:
            self.cancel()


TaskQueue.__doc__ = """TaskQueue is a class that allows you to queue tasks and run them concurrently with a specified number of workers.

    Attributes:
        - `workers` (int): The number of workers to run concurrently. Default is 10.

        - `timeout` (int): The maximum time to wait for the queue to complete. Default is None. If timeout is provided
            the queue is joined using `asyncio.wait_for` with the timeout.

        - `queue` (asyncio.Queue): The queue to store the tasks. Default is `asyncio.PriorityQueue` with no size limit.

        - `on_exit` (Literal["cancel", "complete_priority"]): The action to take when the queue is stopped.

        - `mode` (Literal["finite", "infinite"]): The mode of the queue. If `finite` the queue will stop when all tasks
            are completed. If `infinite` the queue will continue to run until stopped.

        - `worker_timeout` (int): The time to wait for a task to be added to the queue before stopping the worker or
            adding a dummy sleep task to the queue.

        - `stop` (bool): A flag to stop the queue instance.

        - `tasks` (list): A list of the worker tasks running concurrently, including the main task that joins the queue.

        - `priority_tasks` (set): A set to store the QueueItems that must complete before the queue stops.
    """



from .meta_trader import MetaTrader
from .config import Config
from .meta_backtester import MetaBackTester
from .models import *
from .constants import *
from .base import Base, _Base
from .errors import Error
from .exceptions import *
from .task_queue import TaskQueue
from .backtesting import *



from ...lib.candle import Candle, Candles


def find_bearish_fractal(candles: Candles) -> Candle | None:
    """Given a candles object, find the most recent bearish fractal."""
    for i in range(len(candles) - 3, 1, -1):
        if candles[i].high > max(candles[i - 1].high, candles[i + 1].high, candles[i - 2].high, candles[i + 2].high):
            return candles[i]


def find_bullish_fractal(candles: Candles) -> Candle | None:
    """Given a candles object, find the most recent bullish fractal."""
    for i in range(len(candles) - 3, 1, -1):
        if candles[i].low < min(candles[i - 1].low, candles[i + 1].low, candles[i - 2].low, candles[i + 2].low):
            return candles[i]



import asyncio
import os
from concurrent.futures import ThreadPoolExecutor
from signal import signal, SIGINT
from typing import Coroutine, Callable
from logging import getLogger

from ..core.config import Config
from .strategy import Strategy

logger = getLogger(__name__)


class Executor:
    """Executor class for running multiple strategies on multiple symbols concurrently.

    Attributes:
        executor (ThreadPoolExecutor): The executor object.
        strategy_runners (list): List of strategies.
        coroutines (dict[Coroutine, dict]): A list of coroutines to run in the executor
        coroutine_threads (dict[Coroutine, dict]): A list of coroutines to run in the executor
        functions (dict[Callable, dict]): A dictionary of functions to run in the executor
    """
    executor: ThreadPoolExecutor
    config: Config

    def __init__(self):
        self.strategy_runners: list[Strategy] = []
        self.coroutines: dict[Coroutine: dict] = {}
        self.coroutine_threads: dict[Coroutine: dict] = {}
        self.functions: dict[Callable:dict] = {}
        self.config = Config()
        self.timeout = None  # Timeout for the executor. For testing purposes only
        signal(SIGINT, self.sigint_handle)

    def add_function(self, *, function: Callable, kwargs: dict = None):
        kwargs = kwargs or {}
        self.functions[function] = kwargs

    def add_coroutine(self, *, coroutine: Callable | Coroutine, kwargs: dict = None, on_separate_thread=False):
        kwargs = kwargs or {}
        if on_separate_thread:
            self.coroutine_threads[coroutine] = kwargs
        else:
            self.coroutines[coroutine] = kwargs

    def add_strategies(self, *, strategies: tuple[Strategy]):
        """Add multiple strategies at once

        Args:
            strategies (Sequence[Strategy]): A sequence of strategies.
        """
        self.strategy_runners.extend(strategies)

    def add_strategy(self, *, strategy: Strategy):
        """Add a strategy instance to the list of workers

        Args:
            strategy (Strategy): A strategy object
        """
        self.strategy_runners.append(strategy)

    @staticmethod
    def run_strategy(strategy: Strategy):
        """Wraps the coroutine trade method of each strategy with 'asyncio.run'.

        Args:
            strategy (Strategy): A strategy object
        """
        asyncio.run(strategy.run_strategy())

    async def run_coroutine_tasks(self):
        """Run all coroutines in the executor"""
        await asyncio.gather(*[coroutine(**kwargs) for coroutine, kwargs in self.coroutines.items()],
                             return_exceptions=True)

    @staticmethod
    def run_coroutine_task(coroutine, kwargs):
        asyncio.run(coroutine(**kwargs))

    @staticmethod
    def run_function(function: Callable, kwargs: dict):
        """Run a function
        Args:
            function: The function to run
            kwargs: A dictionary of keyword arguments for the function
        """
        function(**kwargs)

    def sigint_handle(self, signum, frame):
        self.config.shutdown = True

    async def exit(self):
        """Shutdown the executor"""
        start = asyncio.get_event_loop().time()
        try:
            while self.config.shutdown is False and self.config.force_shutdown is False:
                if self.timeout is not None and self.timeout < (asyncio.get_event_loop().time() - start):
                    self.config.shutdown = True
                    break

                timeout = 1 if self.timeout else 30
                await asyncio.sleep(timeout)

            for strategy in self.strategy_runners:
                strategy.running = False

            self.config.task_queue.cancel()

            if self.config.backtest_engine is not None:
                self.config.backtest_engine.stop_testing = True

            self.executor.shutdown(wait=False, cancel_futures=False)

            if self.config.force_shutdown:
                os._exit(1)
        except Exception as err:
            logger.error("%s: Unable to shutdown executor", err)
            os._exit(1)

    def execute(self, *, workers: int = 5):
        """Run the strategies with a threadpool executor.

        Args:
            workers: Number of workers to use in executor pool. Defaults to 5.

        Notes:
            No matter the number specified, the executor will always use a minimum of 5 workers.
        """
        workers_ = len(self.strategy_runners) + len(self.functions) + len(self.coroutine_threads) + 3
        workers = max(workers, workers_)
        with ThreadPoolExecutor(max_workers=workers) as executor:
            self.executor = executor
            [self.executor.submit(self.run_strategy, strategy) for strategy in self.strategy_runners]
            [self.executor.submit(function, **kwargs) for function, kwargs in self.functions.items()]
            [self.executor.submit(self.run_coroutine_task, coroutine, kwargs) for coroutine, kwargs in
             self.coroutine_threads.items()]
            self.executor.submit(asyncio.run, self.run_coroutine_tasks())



import asyncio
from datetime import time, timedelta, datetime, UTC
from typing import Literal, Callable, Iterable, NamedTuple
from logging import getLogger

from ..core.models import OrderSendResult, TradePosition
from ..core.config import Config
from .positions import Positions

logger = getLogger(__name__)


class Duration(NamedTuple):
    hours: int
    minutes: int
    seconds: int


def delta(obj: time) -> timedelta:
    """Get the timedelta of a datetime.time object.

    Args:
        obj (datetime.time): A datetime.time object.
    """
    return timedelta(hours=obj.hour, minutes=obj.minute, seconds=obj.second, microseconds=obj.microsecond)


async def backtest_sleep(secs):
    """An async sleep function for use during backtesting."""
    config = Config()
    btc = config.backtest_controller
    sleep = config.backtest_engine.cursor.time + secs
    while sleep > config.backtest_engine.cursor.time:
        btc.wait()


class Session:
    """A session is a time period between two datetime.time objects specified in utc.

    Attributes:
        start (datetime.time): The start time of the session.
        end (datetime.time): The end time of the session.
        on_start (str): The action to take when the session starts. Default is None.
        on_end (str): The action to take when the session ends. Default is None.
        custom_start (Callable): A custom function to call when the session starts. Default is None.
        custom_end (Callable): A custom function to call when the session ends. Default is None.
        name (str): A name for the session. Default is a combination of start and end.
    """

    def __init__(
        self,
        *,
        start: int | time,
        end: int | time,
        on_start: Literal["close_all", "close_win", "close_loss", "custom_start"] = None,
        on_end: Literal["close_all", "close_win", "close_loss", "custom_end"] = None,
        custom_start: Callable = None,
        custom_end: Callable = None,
        name: str = "",
    ):
        """Create a session.

        Keyword Args:
            start (int | datetime.time): The start time of the session in UTC.
            end (int | datetime.time): The end time of the session in UTC.
            on_start (Literal['close_all', 'close_win', 'close_loss', 'custom_start']): The action to take when the
                session starts. Default is None.
            on_end (Literal['close_all', 'close_win', 'close_loss', 'custom_end']): The action to take when the session
                ends. Default is None.
            custom_start (Callable): A custom function to call when the session starts. Default is None.
            custom_end (Callable): A custom function to call when the session ends. Default is None.
            name (str): A name for the session. Default is a combination of start and end.
        """
        self.start = start.replace(tzinfo=UTC) if isinstance(start, time) else time(hour=start, tzinfo=UTC)
        self.end = end if isinstance(end, time) else time(hour=end, tzinfo=UTC)
        self.on_start = on_start
        self.on_end = on_end
        self.custom_start = custom_start
        self.custom_end = custom_end
        self.name = name or f"{self.start}<-->{self.end}"
        self.positions_manager = Positions()
        self.config = Config()

    def __contains__(self, item: time):
        span = (delta(self.end) - delta(self.start)).seconds
        item_span = (delta(self.end) - delta(item)).seconds
        return item_span <= span

    def __str__(self):
        return f"{self.start}<-->{self.end}"

    def __repr__(self):
        return f"{self.start}<-->{self.end}"

    def __len__(self):
        return int((delta(self.end) - delta(self.start)).seconds)

    def in_session(self) -> bool:
        """Check if the current time is within the session."""
        now = (
            datetime.now(tz=UTC).time()
            if self.config.mode == "live"
            else datetime.fromtimestamp(self.config.backtest_engine.cursor.time, tz=UTC).time()
        )
        return now in self

    async def begin(self):
        """Call the action specified in on_start or custom_start."""
        await self.action(action=self.on_start)

    async def close(self):
        """Call the action specified in on_end or custom_end."""
        await self.action(action=self.on_end)

    def duration(self) -> Duration:
        """Get the duration of the session in seconds."""
        hours, seconds = divmod(len(self), 3600)
        minutes, seconds = divmod(seconds, 60)
        return Duration(hours=hours, minutes=minutes, seconds=seconds)

    async def close_positions(self, *, positions: tuple[TradePosition, ...]):
        results = asyncio.gather(
            *(self.positions_manager.close_position(position=position) for position in positions),
            return_exceptions=True,
        )
        closed = pending = 0
        for result in results:
            if isinstance(result, OrderSendResult) and result.retcode == 10009:
                closed += 1
                continue
            pending += 1
        logger.info(f"Closed {closed} positions")
        logger.warning(f"{pending} positions still pending") if pending else ...

    async def close_all(self):
        open_positions = await self.positions_manager.get_positions()
        await self.close_positions(positions=open_positions)

    async def close_win(self):
        open_positions = await self.positions_manager.get_positions()
        positions = tuple(position for position in open_positions if position.profit >= 0)
        await self.close_positions(positions=positions)

    async def close_loss(self):
        open_positions = await self.positions_manager.get_positions()
        positions = tuple(position for position in open_positions if position.profit < 0)
        await self.close_positions(positions=positions)

    async def action(self, *, action):
        """Used by begin and close to call the action specified.

        Args:
            action (Literal['close_all', 'close_win', 'close_loss', 'custom_start', 'custom_end']): The action to take.
        """
        try:
            match action:
                case "close_all":
                    await self.close_all()

                case "close_win":
                    await self.close_win()

                case "close_loss":
                    await self.close_loss()

                case "custom_end":
                    await self.custom_end()

                case "custom_start":
                    await self.custom_start()

                case _:
                    pass
        except Exception as exe:
            logger.warning(f"Failed to call action {action} due to {exe}")

    def until(self):
        """Get the seconds until the session starts from the current time in seconds."""
        if self.config.mode == "backtest":
            now = datetime.fromtimestamp(self.config.backtest_engine.cursor.time, tz=UTC).time()
            secs = (delta(self.start) - delta(now)).seconds
        else:
            secs = (delta(self.start) - delta(datetime.now(tz=UTC).time())).seconds
        return secs


class Sessions:
    """Sessions allow you to run code at specific times of the day. It is a collection of Session objects.
    Sessions are sorted by start time. The sessions object is an asynchronous context manager.

    Attributes:
        sessions (list[Session]): A list of Session objects.
        current_session (Session): The current session.
    """

    sessions: list[Session]
    current_session: Session | None

    def __init__(self, *, sessions: Iterable[Session]):
        self.sessions = list(sessions)
        self.sessions.sort(key=lambda x: (x.start.hour, x.end.hour))
        self.current_session = None
        self.config = Config()

    def find(self, *, moment: time = None) -> Session | None:
        """Find a session that contains a datetime.time object, if not found return None.

        Keyword Args:
            moment (datetime.time | None): A datetime.time object. if not provided, the current time is used.

        Returns:
            Session | None: A Session object or None if not found.
        """
        moment = (
            moment or datetime.now(tz=UTC).time()
            if self.config.mode == "live"
            else (datetime.fromtimestamp(self.config.backtest_engine.cursor.time, tz=UTC).time())
        )
        for session in self.sessions:
            if moment in session:
                return session
        return None

    def find_next(self, *, moment: time = None) -> Session:
        """Find the next session that contains a datetime.time object.

        Args:
            moment (datetime.time | None): A datetime.time object, if not provided, the current time is used.

        Returns:
            Session: A Session object.
        """
        moment = (
            moment or datetime.now(tz=UTC).time()
            if self.config.mode == "live"
            else (datetime.fromtimestamp(self.config.backtest_engine.cursor.time, tz=UTC).time())
        )
        for session in self.sessions:
            if delta(moment) < delta(session.start):
                return session
        return self.sessions[0]

    def __contains__(self, moment: time):
        return True if self.find(moment=moment) is not None else False

    async def __aenter__(self):
        await self.check()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.current_session.close() if self.current_session is not None else ...

    async def check(self):
        """Check if the current session has started and if not, wait until it starts."""
        if self.current_session is not None and self.current_session.in_session():
            return

        if self.config.mode == "backtest":
            now = datetime.fromtimestamp(self.config.backtest_engine.cursor.time, tz=UTC).time()
        else:
            now = datetime.now(tz=UTC).time()

        next_session = self.find(moment=now)

        if next_session and self.current_session is None:
            self.current_session = next_session
            await self.current_session.begin()
            return

        if next_session and self.current_session is not None:
            await self.current_session.close()
            self.current_session = next_session
            await self.current_session.begin()
            return

        if next_session is None and self.current_session is not None:
            await self.current_session.close()

        next_session = self.find_next(moment=now)
        secs = next_session.until() + 10
        logger.info(f"sleeping for {secs} seconds until next {next_session} session")
        sleep_func = asyncio.sleep if self.config.mode == "live" else backtest_sleep
        await sleep_func(secs)
        self.current_session = next_session
        await self.current_session.begin()



from logging import getLogger

from ...core.models import OrderType
from ...lib.trader import Trader

logger = getLogger(__name__)


class SimpleTrader(Trader):
    async def place_trade(self, *, order_type: OrderType, sl: float, parameters: dict = None):
        """Places a trade based on the order_type and a given stop_loss. The volume is based on the amount to risk which is
           calculated using the Risk Assessment Management instance.

        Args:
            order_type (OrderType): The order_type
            sl (float): The stop_loss
            parameters (dict): Parameters associated with the trade
        """
        try:
            self.parameters |= parameters or {}
            await self.create_order_with_sl(order_type=order_type, sl=sl)
            if not await self.check_order():
                return
            self.order.comment = self.parameters.get("name", self.__class__.__name__)
            await self.send_order()
        except Exception as err:
            logger.error(f"{err} in {self.__class__.__name__}.place_trade for {self.symbol.name}")



from datetime import datetime
from typing import TypeVar, Generic
from logging import getLogger

from MetaTrader5 import TradePosition, TradeOrder, TradeDeal

logger = getLogger(__name__)

TradeData = TypeVar("TradeData", bound=TradePosition | TradeOrder | TradeDeal)


class TradeManager(Generic[TradeData]):
    """A generic class to manage trades data during a backtest. It is the parent class of the
     PositionsManager, OrdersManager, and DealsManager. It implements some dict-like methods to manage the data.
     It has a private attribute _data to store the data. It exposes the data through the values, keys, and items methods.
    It also has a to_dict method to convert the data to a dictionary.

    Attributes:
        _data (dict[int, TradeData]): The data to store the trades.

    Examples:
        >>> manager = TradeManager()
        >>> manager[123456] = TradePosition(ticket=123456, symbol="EURUSD", volume=0.1)
        >>> manager.update(ticket=123456, symbol="EURUSD", volume=0.1)
        >>> manager[123456]
        TradePosition(ticket=123456, symbol='EURUSD', volume=0.1)
        >>> manager.values()
        (TradePosition(ticket=123456, symbol='EURUSD', volume=0.1),)
        >>> manager.keys()
        (123456,)
        >>> manager.items()
        ((123456, TradePosition(ticket=123456, symbol='EURUSD', volume=0.1)),)
        >>> manager.to_dict()
        {123456: {'ticket': 123456, 'symbol': 'EURUSD', 'volume': 0.1}}
        >>> pos = manager.get(123456)
        >>> pos
        TradePosition(ticket=123456, symbol='EURUSD', volume=0.1)
        >>> pos in manager
        True
        >>> len(manager)
        1
        >>> pos in manager
        False
    """

    _data: dict[int, TradeData]

    def __init__(self, *, data: dict = None):
        self._data = data or {}

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __contains__(self, item: TradeData):
        return item.ticket in self._data

    def __getitem__(self, item):
        return self._data[item]

    def __setitem__(self, key, value: TradeData):
        self._data[key] = value

    def __delitem__(self, key):
        del self._data[key]

    def get(self, key, default=None) -> TradeData | None:
        return self._data.get(key, default)

    def update(self, *, ticket: int, **kwargs):
        """Update the data of a trade. Given the ticket of the trade and the new data to update.

        Args:
            ticket (int): The ticket of the trade to update.
            **kwargs: The new data to update.
        """
        try:
            res = self[ticket]
            klass = type(res)
            res = res._asdict()
            res.update(**kwargs)
            res = klass(res.get(v) for v in klass.__match_args__)
            self[res.ticket] = res
            return res
        except KeyError:
            logger.error(f"Update Operation Failed: Could Not Find Ticket")

    def values(self) -> tuple[TradeData, ...]:
        """Returns the values of the data."""
        return tuple(value for value in self._data.values())

    def keys(self) -> tuple[int, ...]:
        """Returns the keys of the data."""
        return tuple(key for key in self._data.keys())

    def items(self) -> tuple[tuple[int, TradeData], ...]:
        """Returns the items of the data."""
        return tuple((key, value) for key, value in self._data.items())

    def to_dict(self):
        """Convert the data to a dictionary."""
        return {key: value._asdict() for key, value in self._data.items()}


class PositionsManager(TradeManager):
    """A class to manage the open positions during a backtest. It is a subclass of TradeManager. It has an additional
    attribute _open_positions to store the open positions. It also has a margins attribute to store the margins of the
    open positions. It overrides some methods of the TradeManager class to manage the open positions.

    Attributes:
        _open_positions (set[int]): The open positions.
        margins (dict[int, float]): The margins of the open positions.
    """

    _data: dict[int, TradePosition]
    _open_positions: set[int]
    margins: dict[int, float]

    def __init__(self, *, data: dict = None, open_positions: set[int] = None, margins: dict = None):
        """Positions manager manages the open positions during a backtest. It is a subclass of TradeManager. It has an
        additional attribute _open_positions to store the open positions. It also has a margins attribute to store the
        margins of the open positions. It overrides some methods of the TradeManager class to manage the open positions.

        Args:
            data (dict, optional): The data to store the trades. This used for continuation of the backtesting, if it
             was stopped with some open positions.

            open_positions (set, optional): The open positions. Defaults to None.

            margins (dict, optional): The margins of the open positions. Defaults to None.
        """
        super().__init__(data=data)
        self._open_positions = open_positions or {trade.ticket for trade in self._data.values()}
        self.margins: dict[int, float] = margins or dict()

    def __len__(self):
        return len(self._open_positions)

    def __contains__(self, item: TradePosition):
        return item.ticket in self._open_positions

    def __getitem__(self, item):
        if item in self._open_positions:
            return super().__getitem__(item)
        raise KeyError("Position not found")

    def __setitem__(self, key, value: TradeData):
        self._open_positions.add(value.ticket)
        self._data[key] = value

    def __delitem__(self, key):
        self._open_positions.discard(key)
        del self._data[key]

    @property
    def margin(self):
        """Returns the total margin of all open positions"""
        return sum(self.margins.values())

    def close(self, *, ticket: int) -> bool:
        """Close a position. Given the ticket of the position to close.

        Args:
            ticket (int): The ticket of the position to close.
        """
        is_open = ticket in self._open_positions
        self._open_positions.discard(ticket)
        return is_open

    def get_margin(self, *, ticket: int) -> float:
        """Get the margin of a position. Given the ticket of the position.

        Args:
            ticket (int): The ticket of the position.

        Returns:
            float: The margin of the position.
        """
        return self.margins.get(ticket, 0.0)

    def delete_margin(self, *, ticket: int):
        """Delete the margin of a position. Given the ticket of the position.

        Args:
            ticket (int): The ticket of the position.
        """
        return self.margins.pop(ticket, 0)

    def set_margin(self, *, ticket: int, margin: float):
        """Set the margin of a position. Given the ticket of the position and the margin.

        Args:
            ticket (int): The ticket of the position.
            margin (float): The margin of the position
        """
        self.margins[ticket] = margin

    def positions_get(self, *, ticket: int = None, symbol: str = None, group: None = None) -> tuple[TradePosition, ...]:
        """Get positions. Given the ticket, symbol, or group of the positions.

        Args:
            ticket (int): The ticket of the position.
            symbol (str): The symbol of the position.
            group (str): The group

        Returns:
            tuple[TradePosition, ...]: The positions
        """
        if ticket:
            return tuple(position for position in self.open_positions if position.ticket == ticket)

        if symbol:
            return tuple(position for position in self.open_positions if position.symbol == symbol)

        if group:
            return self.open_positions

        if ticket == group == symbol is None:
            return self.open_positions

        return tuple()

    def positions_total(self) -> int:
        """Get the total number of open positions.

        Returns:
            int: The total number of open positions.
        """
        return len(self._open_positions)

    @property
    def open_positions(self) -> tuple[TradePosition, ...]:
        """Returns the open positions.

        Args:
            tuple[TradePosition, ...]: The open positions.
        """
        return tuple(position for position in self.values() if position.ticket in self._open_positions)


class OrdersManager(TradeManager):
    """Managers orders data during a backtest. It is a subclass of TradeManager. It manages access to the historical
    orders data
    """

    _data = dict[int, TradeOrder]

    def get_orders_range(self, *, date_from: float, date_to: float) -> tuple[TradeData, ...]:
        """Get orders within a date range. Given the start and end date of the range.

        Args:
            date_from (float): The start date of the range.
            date_to (float): The end date of the range.

        Returns:
            tuple[TradeData, ...]: The orders within the date range.
        """
        start = date_from.timestamp() if isinstance(date_from, datetime) else date_from
        end = date_to.timestamp() if isinstance(date_to, datetime) else date_to
        return tuple(order for order in self.values() if start <= order.time_setup <= end)

    def history_orders_get(
        self,
        *,
        date_from: float | datetime = None,
        date_to: float | datetime = None,
        group: str = "",
        ticket: int = None,
        position: int = None,
    ) -> tuple[TradeOrder, ...]:
        """Get historical orders. Given the start and end date of the range, the group, ticket, or position of the
         orders.

         Args:
            date_from (float, datetime): The start date of the range.
            date_to (float, datetime): The end date of the range.
            group (str): The group of the orders.
            ticket (int): The ticket of the order.
            position (int): The position of the order.

        Returns:
            tuple[TradeOrder, ...]: The historical orders.
        """
        if date_from and date_to:
            orders = self.get_orders_range(date_from=date_from, date_to=date_to)
            if group:
                orders = orders
            return orders

        if ticket:
            return tuple(order for order in self.values() if order.ticket == ticket)

        if position:
            return tuple(order for order in self.values() if order.position_id == position)

        return ()

    def history_orders_total(self, *, date_from: datetime | float, date_to: datetime | float) -> int:
        """Get the total number of historical orders. Given the start and end date of the range.

        Args:
            date_from (datetime, float): The start date of the range.
            date_to (datetime, float): The end date of the range.
        """
        return len(self.get_orders_range(date_from=date_from, date_to=date_to))


class DealsManager(TradeManager):
    _data = dict[int, TradeDeal]

    def get_deals_range(self, *, date_from: float, date_to: float) -> tuple[TradeData, ...]:
        """Get deals within a date range. Given the start and end date of the range.

        Args:
            date_from (float): The start date of the range.
            date_to (float): The end date of the range.

        Returns:
            tuple[TradeData, ...]: The deals within the date range.
        """
        start = date_from.timestamp() if isinstance(date_from, datetime) else date_from
        end = date_to.timestamp() if isinstance(date_to, datetime) else date_to
        return tuple(deal for deal in self.values() if start <= deal.time <= end)

    def history_deals_get(
        self,
        *,
        date_from: float | datetime = None,
        date_to: float | datetime = None,
        group: str = "",
        ticket: int = None,
        position: int = None,
    ) -> tuple[TradeDeal, ...]:
        """History deals get. Given the start and end date of the range, the group, ticket, or position of the deals.

        Args:
            date_from (float, datetime): The start date of the range.
            date_to (float, datetime): The end date of the range.
            group (str): The group of the deals.
            ticket (int): The ticket of the deal.
            position (int): The position of the deal.
        """
        if date_from and date_to:
            deals = self.get_deals_range(date_from=date_from, date_to=date_to)
            if group:
                deals = deals
            return deals

        if ticket:
            return tuple(deal for deal in self.values() if deal.ticket == ticket)

        if position and (ticket is None):
            return tuple(deal for deal in self.values() if deal.position_id == position)

        return ()

    def history_deals_total(self, *, date_from: datetime | float, date_to: datetime | float) -> int:
        """Get the total number of historical deals. Given the start and end date of the range

        Args:
            date_from (datetime, float): The start date of the range.
            date_to (datetime, float): The end date of the range.

        Returns:
            int: The total number of historical deals.
        """
        return len(self.get_deals_range(date_from=date_from, date_to=date_to))



"""Risk Assessment and Management"""

from .account import Account
from .positions import Positions


class RAM:
    account: Account
    risk_to_reward: float
    risk: float
    fixed_amount: float | None
    min_amount: float
    max_amount: float
    loss_limit: int
    open_limit: int

    def __init__(self, **kwargs):
        """Initialize Risk Assessment and Management with the provided keyword arguments.

        Keyword Args:
            risk_to_reward (float): Risk to reward ratio. Defaults to 2
            risk (float): Percentage of capital to risk per trade. Defaults to 1%
            min_amount (float): Minimum amount to risk per trade. Defaults to 0
            max_amount (float): Maximum amount to risk per trade. Defaults to 0
            loss_limit (int): Maximum number of losing positions. Defaults to 1
            open_limit (int): Maximum number of open positions. Defaults to 1
            fixed_amount (float): Fixed amount to risk per trade. Defaults to None
        """
        self.account = Account()
        self.positions = Positions()
        self.risk_to_reward = kwargs.get("risk_to_reward", 2)
        self.risk = kwargs.get("risk", 1)
        self.min_amount = kwargs.get("min_amount", 0)
        self.max_amount = kwargs.get("max_amount", 0)
        self.loss_limit = kwargs.get("loss_limit", 3)
        self.open_limit = kwargs.get("open_limit", 3)
        self.fixed_amount = kwargs.get("fixed_amount", None)

    def modify_ram(self, **kwargs):
        """Modify the Risk Assessment and Management with the provided keyword arguments.
        """
        [setattr(self, key, value) for key, value in kwargs.items()]

    async def get_amount(self) -> float:
        """Calculate the amount to risk per trade as a percentage of margin_free.

        Returns:
            float: Amount to risk per trade
        """
        if self.fixed_amount:
            return self.fixed_amount
        await self.account.refresh()
        amount = self.account.margin_free * (self.risk / 100)
        if self.min_amount and self.max_amount:
            return max(self.min_amount, min(self.max_amount, amount))
        return amount

    async def check_losing_positions(self) -> bool:
        """Check if the number of losing positions is less than the loss limit

        Returns:
            bool: True if the number of losing positions is less than or equal the loss limit
        """
        positions = await self.positions.get_positions()
        loosing = [position for position in positions if position.profit < 0]
        return len(loosing) <= self.loss_limit

    async def check_open_positions(self) -> bool:
        """Check if the number of open positions is less than or equal the loss limit.

        Returns:
            bool: True if the number of open positions is less than the open limit
        """
        positions = await self.positions.get_positions()
        return len(positions) <= self.open_limit



from logging import getLogger

from ..core.models import TradeRequest, TradeOrder, OrderCheckResult, OrderSendResult
from ..core.constants import TradeAction, OrderTime, OrderFilling
from ..core.exceptions import OrderError
from ..core.base import _Base
from .._utils import backoff_decorator, error_handler

logger = getLogger(__name__)


class Order(_Base, TradeRequest):
    """Trade order related functions and properties. Subclass of TradeRequest."""

    def __init__(self, **kwargs):
        """Initialize the order object with keyword arguments, symbol must be provided.
        Provide default values for action, type_time and type_filling if not provided.

        Args:
            **kwargs: Keyword arguments must match the attributes of TradeRequest as well as the attributes of
             Order class as specified in the annotations in the class definition.

        Default Values:
            action (TradeAction.DEAL): Trade action
            type_time (OrderTime.DAY): Order time
            type_filling (OrderFilling.FOK): Order filling
        """
        kwargs = {"action": TradeAction.DEAL, "type_time": OrderTime.DAY, "type_filling": OrderFilling.FOK, **kwargs}
        super().__init__(**kwargs)

    def modify(self, **kwargs):
        """Modify the order object with keyword arguments.

        Args:
            **kwargs: Keyword arguments must match the attributes of TradeRequest as well as the attributes of
             Order class as specified in the annotations in the class definition.
        """
        self.set_attributes(**kwargs)

    async def orders_total(self):
        """Get the number of active pending orders.

        Returns:
            (int): total number of active pending orders
        """
        return await self.mt5.orders_total()

    async def get_pending_order(self, *, ticket: int) -> TradeOrder | None:
        """
        Get a pending order by ticket number.

        Args:
            ticket (int): Order ticket number

        Returns:
        """
        orders = await self.mt5.orders_get(ticket=ticket)
        order = None
        for order_ in orders:
            if order_.ticket == ticket:
                return TradeOrder(**order_._asdict())
        return order

    async def get_pending_orders(self, *, ticket: int = 0, symbol: str = "", group: str = "") -> tuple[TradeOrder, ...]:
        """Get the list of active pending orders for the current symbol.

        Args:
            ticket (int): Order ticket number
            symbol (str): Symbol name
            group (str): Group name

        Returns:
            tuple[TradeOrder, ...]: A Tuple of active pending trade orders as TradeOrder objects
        """
        orders = await self.mt5.orders_get(symbol=symbol, ticket=ticket, group=group)
        if orders is not None:
            return tuple(TradeOrder(**order._asdict()) for order in orders)
        return tuple()

    @backoff_decorator
    async def check(self, **kwargs) -> OrderCheckResult:
        """Check funds sufficiency for performing a required trading operation and the possibility of executing it.

        Returns:
            OrderCheckResult: An OrderCheckResult object

        Raises:
            OrderError: If not successful
        """
        req = self.request | kwargs
        res = await self.mt5.order_check(req)
        if res is None:
            raise OrderError(f"Order check failed for {self.symbol}")
        return OrderCheckResult(**res._asdict())

    @backoff_decorator
    async def send(self) -> OrderSendResult:
        """Send a request to perform a trading operation from the terminal to the trade server.

        Returns:
             OrderSendResult: An OrderSendResult object

        Raises:
            OrderError: If not successful
        """
        res = await self.mt5.order_send(self.request)
        if res is None:
            raise OrderError(f"Failed to send order {self.symbol}")
        return OrderSendResult(**res._asdict())

    @error_handler(log_error_msg=False)
    async def calc_margin(self) -> float | None:
        """Return the required margin in the account currency to perform a specified trading operation.

        Returns:
            float: Returns float value if successful
        """
        res = await self.mt5.order_calc_margin(self.type, self.symbol, self.volume, self.price)
        return res

    @error_handler(response=0, log_error_msg=False)
    async def calc_profit(self) -> float:
        """Return profit in the account currency for a specified trading operation.

        Returns:
            float: Returns float value if successful
            None: If not successful
        """
        action, symbol, volume, price_open, price_close = (self.type, self.symbol, self.volume, self.price, self.tp)
        res = await self.mt5.order_calc_profit(action, symbol, volume, price_open, price_close)
        return res

    @error_handler(response=0, log_error_msg=False)
    async def calc_loss(self) -> float:
        """Return profit in the account currency for a specified trading operation.

        Returns:
            float: Returns float value if successful
            None: If not successful
        """
        action, symbol, volume, price_open, price_close = (self.type, self.symbol, self.volume, self.price, self.sl)
        res = await self.mt5.order_calc_profit(action, symbol, volume, price_open, price_close)
        return res

    @property
    def request(self) -> dict:
        """Return the order request as a dictionary."""
        return {key: value for key, value in self.dict.items() if key in self.mt5.TradeRequest.__match_args__}



import logging

from ...lib.symbol import Symbol
from ...lib.trader import Trader
from ...lib.candle import Candles
from ...lib.strategy import Strategy
from ...core.constants import TimeFrame, OrderType
from ...lib.sessions import Sessions
from ..candle_patterns import find_bearish_fractal, find_bullish_fractal
from ..traders import SimpleTrader
from ..utils.tracker import Tracker

logger = logging.getLogger(__name__)


class FingerTrap(Strategy):
    ttf: TimeFrame
    etf: TimeFrame
    fast_ema: int
    slow_ema: int
    entry_ema: int
    ecc: int
    tcc: int
    trader: Trader
    tracker: Tracker

    # The default parameters for the strategy. You can override these in the constructor.
    # via the `params` argument.
    parameters = {"fast_ema": 8, "slow_ema": 20, "etf": TimeFrame.M5, "ttf": TimeFrame.H1,
                  "entry_ema": 5, "tcc": 720, "ecc": 8640}

    def __init__(self, *,symbol: Symbol, params: dict | None = None, trader: Trader = None, sessions: Sessions = None,
                 name: str = "FingerTrap"):
        super().__init__(symbol=symbol, params=params, sessions=sessions, name=name)
        self.trader = trader or SimpleTrader(symbol=self.symbol)
        self.tracker: Tracker = Tracker(snooze=self.ttf.seconds)

    async def check_trend(self):
        try:
            candles: Candles = await self.symbol.copy_rates_from_pos(timeframe=self.ttf, count=self.tcc)
            if (current := candles[-1]) and current.time < self.tracker.trend_time:
                self.tracker.update(new=False, order_type=None)
                return

            self.tracker.update(new=True, trend_time=current.time, last_trend_price=current.close)
            candles.ta.ema(length=self.slow_ema, append=True, fillna=0)
            candles.ta.ema(length=self.fast_ema, append=True, fillna=0)
            candles.rename(inplace=True, **{f"EMA_{self.fast_ema}": "fast", f"EMA_{self.slow_ema}": "slow"})

            fas = candles.ta_lib.above(candles.fast, candles.slow)
            fbs = candles.ta_lib.below(candles.fast, candles.slow)
            caf = candles.ta_lib.above(candles.close, candles.fast)
            cbf = candles.ta_lib.below(candles.close, candles.fast)

            if fas.iloc[-1] and caf.iloc[-1]:
                self.tracker.update(trend="bullish")

            elif fbs.iloc[-1] and cbf.iloc[-1]:
                self.tracker.update(trend="bearish")

            else:
                self.tracker.update(trend="ranging", snooze=self.ttf.seconds, order_type=None)
        except Exception as err:
            logger.error(f"{err} for {self.symbol} in {self.__class__.__name__}.check_trend")
            self.tracker.update(snooze=self.ttf.seconds, order_type=None)

    async def confirm_trend(self):
        try:
            candles = await self.symbol.copy_rates_from_pos(timeframe=self.etf, count=self.ecc)
            if (current := candles[-1]) and current.time < self.tracker.entry_time:
                self.tracker.update(new=False, order_type=None)
                return

            self.tracker.update(new=True, entry_time=current.time, last_entry_price=current.close)
            candles.ta.ema(length=self.entry_ema, append=True)
            candles.rename(**{f"EMA_{self.entry_ema}": "ema"})
            candles["cae"] = candles.ta_lib.cross(candles.close, candles.ema)
            candles["cbe"] = candles.ta_lib.cross(candles.close, candles.ema, above=False)
            current = candles[-1]

            if self.tracker.bullish and current.cae:
                sl = find_bullish_fractal(candles).low
                self.tracker.update(snooze=self.ttf.seconds, order_type=OrderType.BUY, sl=sl)
            elif self.tracker.bearish and current.cbe:
                sl = find_bearish_fractal(candles).high
                self.tracker.update(snooze=self.ttf.seconds, order_type=OrderType.SELL, sl=sl)
            else:
                self.tracker.update(snooze=self.etf.seconds, order_type=None)
        except Exception as err:
            logger.error(f"{err} for {self.symbol} in {self.__class__.__name__}.confirm_trend")
            self.tracker.update(snooze=self.etf.seconds, order_type=None)

    async def watch_market(self):
        await self.check_trend()
        if self.tracker.ranging is False:
            await self.confirm_trend()

    async def trade(self):
        try:
            await self.watch_market()
            if self.tracker.new is False:
                await self.sleep(secs=5)
            elif self.tracker.order_type is None:
                await self.sleep(secs=self.tracker.snooze)
            else:
                await self.trader.place_trade(order_type=self.tracker.order_type, parameters=self.parameters,
                                              sl=self.tracker.sl)
            await self.sleep(secs=self.tracker.snooze)
        except Exception as err:
            logger.error(f"{err} For {self.symbol} in {self.__class__.__name__}.trade")
            await self.sleep(secs=self.ttf.seconds)



from typing import Callable

import MetaTrader5
from MetaTrader5 import (
    Tick,
    SymbolInfo,
    AccountInfo,
    TerminalInfo,
    TradeOrder,
    TradePosition,
    TradeDeal,
    OrderCheckResult,
    OrderSendResult,
    BookInfo,
    TradeRequest,
)

from .config import Config

constants = (
    "TIMEFRAME_M1",
    "TIMEFRAME_M2",
    "TIMEFRAME_M3",
    "TIMEFRAME_M4",
    "TIMEFRAME_M5",
    "TIMEFRAME_M6",
    "TIMEFRAME_M10",
    "TIMEFRAME_M12",
    "TIMEFRAME_M15",
    "TIMEFRAME_M20",
    "TIMEFRAME_M30",
    "TIMEFRAME_H1",
    "TIMEFRAME_H2",
    "TIMEFRAME_H4",
    "TIMEFRAME_H3",
    "TIMEFRAME_H6",
    "TIMEFRAME_H8",
    "TIMEFRAME_H12",
    "TIMEFRAME_D1",
    "TIMEFRAME_W1",
    "TIMEFRAME_MN1",
    "COPY_TICKS_ALL",
    "COPY_TICKS_INFO",
    "COPY_TICKS_TRADE",
    "TICK_FLAG_BID",
    "TICK_FLAG_ASK",
    "TICK_FLAG_LAST",
    "TICK_FLAG_VOLUME",
    "TICK_FLAG_BUY",
    "TICK_FLAG_SELL",
    "POSITION_TYPE_BUY",
    "POSITION_TYPE_SELL",
    "POSITION_REASON_CLIENT",
    "POSITION_REASON_MOBILE",
    "POSITION_REASON_WEB",
    "POSITION_REASON_EXPERT",
    "ORDER_TYPE_BUY",
    "ORDER_TYPE_SELL",
    "ORDER_TYPE_BUY_LIMIT",
    "ORDER_TYPE_SELL_LIMIT",
    "ORDER_TYPE_BUY_STOP",
    "ORDER_TYPE_SELL_STOP",
    "ORDER_TYPE_BUY_STOP_LIMIT",
    "ORDER_TYPE_SELL_STOP_LIMIT",
    "ORDER_TYPE_CLOSE_BY",
    "ORDER_STATE_STARTED",
    "ORDER_STATE_PLACED",
    "ORDER_STATE_CANCELED",
    "ORDER_STATE_PARTIAL",
    "ORDER_STATE_FILLED",
    "ORDER_STATE_REJECTED",
    "ORDER_STATE_EXPIRED",
    "ORDER_STATE_REQUEST_ADD",
    "ORDER_STATE_REQUEST_MODIFY",
    "ORDER_STATE_REQUEST_CANCEL",
    "ORDER_FILLING_FOK",
    "ORDER_FILLING_IOC",
    "ORDER_FILLING_RETURN",
    "ORDER_FILLING_BOC",
    "ORDER_TIME_GTC",
    "ORDER_TIME_DAY",
    "ORDER_TIME_SPECIFIED",
    "ORDER_TIME_SPECIFIED_DAY",
    "ORDER_REASON_CLIENT",
    "ORDER_REASON_MOBILE",
    "ORDER_REASON_WEB",
    "ORDER_REASON_EXPERT",
    "ORDER_REASON_SL",
    "ORDER_REASON_TP",
    "ORDER_REASON_SO",
    "DEAL_TYPE_BUY",
    "DEAL_TYPE_SELL",
    "DEAL_TYPE_BALANCE",
    "DEAL_TYPE_CREDIT",
    "DEAL_TYPE_CHARGE",
    "DEAL_TYPE_CORRECTION",
    "DEAL_TYPE_BONUS",
    "DEAL_TYPE_COMMISSION",
    "DEAL_TYPE_COMMISSION_DAILY",
    "DEAL_TYPE_COMMISSION_MONTHLY",
    "DEAL_TYPE_COMMISSION_AGENT_DAILY",
    "DEAL_TYPE_COMMISSION_AGENT_MONTHLY",
    "DEAL_TYPE_INTEREST",
    "DEAL_TYPE_BUY_CANCELED",
    "DEAL_TYPE_SELL_CANCELED",
    "DEAL_DIVIDEND",
    "DEAL_DIVIDEND_FRANKED",
    "DEAL_TAX",
    "DEAL_ENTRY_IN",
    "DEAL_ENTRY_OUT",
    "DEAL_ENTRY_INOUT",
    "DEAL_ENTRY_OUT_BY",
    "DEAL_REASON_CLIENT",
    "DEAL_REASON_MOBILE",
    "DEAL_REASON_WEB",
    "DEAL_REASON_EXPERT",
    "DEAL_REASON_SL",
    "DEAL_REASON_TP",
    "DEAL_REASON_SO",
    "DEAL_REASON_ROLLOVER",
    "DEAL_REASON_VMARGIN",
    "DEAL_REASON_SPLIT",
    "TRADE_ACTION_DEAL",
    "TRADE_ACTION_PENDING",
    "TRADE_ACTION_SLTP",
    "TRADE_ACTION_MODIFY",
    "TRADE_ACTION_REMOVE",
    "TRADE_ACTION_CLOSE_BY",
    "SYMBOL_CHART_MODE_BID",
    "SYMBOL_CHART_MODE_LAST",
    "SYMBOL_CALC_MODE_FOREX",
    "SYMBOL_CALC_MODE_FUTURES",
    "SYMBOL_CALC_MODE_CFD",
    "SYMBOL_CALC_MODE_CFDINDEX",
    "SYMBOL_CALC_MODE_CFDLEVERAGE",
    "SYMBOL_CALC_MODE_FOREX_NO_LEVERAGE",
    "SYMBOL_CALC_MODE_EXCH_STOCKS",
    "SYMBOL_CALC_MODE_EXCH_FUTURES",
    "SYMBOL_CALC_MODE_EXCH_OPTIONS",
    "SYMBOL_CALC_MODE_EXCH_OPTIONS_MARGIN",
    "SYMBOL_CALC_MODE_EXCH_BONDS",
    "SYMBOL_CALC_MODE_EXCH_STOCKS_MOEX",
    "SYMBOL_CALC_MODE_EXCH_BONDS_MOEX",
    "SYMBOL_CALC_MODE_SERV_COLLATERAL",
    "SYMBOL_TRADE_MODE_DISABLED",
    "SYMBOL_TRADE_MODE_LONGONLY",
    "SYMBOL_TRADE_MODE_SHORTONLY",
    "SYMBOL_TRADE_MODE_CLOSEONLY",
    "SYMBOL_TRADE_MODE_FULL",
    "SYMBOL_TRADE_EXECUTION_REQUEST",
    "SYMBOL_TRADE_EXECUTION_INSTANT",
    "SYMBOL_TRADE_EXECUTION_MARKET",
    "SYMBOL_TRADE_EXECUTION_EXCHANGE",
    "SYMBOL_SWAP_MODE_DISABLED",
    "SYMBOL_SWAP_MODE_POINTS",
    "SYMBOL_SWAP_MODE_CURRENCY_SYMBOL",
    "SYMBOL_SWAP_MODE_CURRENCY_MARGIN",
    "SYMBOL_SWAP_MODE_CURRENCY_DEPOSIT",
    "SYMBOL_SWAP_MODE_INTEREST_CURRENT",
    "SYMBOL_SWAP_MODE_INTEREST_OPEN",
    "SYMBOL_SWAP_MODE_REOPEN_CURRENT",
    "SYMBOL_SWAP_MODE_REOPEN_BID",
    "DAY_OF_WEEK_SUNDAY",
    "DAY_OF_WEEK_MONDAY",
    "DAY_OF_WEEK_TUESDAY",
    "DAY_OF_WEEK_WEDNESDAY",
    "DAY_OF_WEEK_THURSDAY",
    "DAY_OF_WEEK_FRIDAY",
    "DAY_OF_WEEK_SATURDAY",
    "SYMBOL_ORDERS_GTC",
    "SYMBOL_ORDERS_DAILY",
    "SYMBOL_ORDERS_DAILY_NO_STOPS",
    "SYMBOL_OPTION_RIGHT_CALL",
    "SYMBOL_OPTION_RIGHT_PUT",
    "SYMBOL_OPTION_MODE_EUROPEAN",
    "SYMBOL_OPTION_MODE_AMERICAN",
    "ACCOUNT_TRADE_MODE_DEMO",
    "ACCOUNT_TRADE_MODE_CONTEST",
    "ACCOUNT_TRADE_MODE_REAL",
    "ACCOUNT_STOPOUT_MODE_PERCENT",
    "ACCOUNT_STOPOUT_MODE_MONEY",
    "ACCOUNT_MARGIN_MODE_RETAIL_NETTING",
    "ACCOUNT_MARGIN_MODE_EXCHANGE",
    "ACCOUNT_MARGIN_MODE_RETAIL_HEDGING",
    "BOOK_TYPE_SELL",
    "BOOK_TYPE_BUY",
    "BOOK_TYPE_SELL_MARKET",
    "BOOK_TYPE_BUY_MARKET",
    "TRADE_RETCODE_REQUOTE",
    "TRADE_RETCODE_REJECT",
    "TRADE_RETCODE_CANCEL",
    "TRADE_RETCODE_PLACED",
    "TRADE_RETCODE_DONE",
    "TRADE_RETCODE_DONE_PARTIAL",
    "TRADE_RETCODE_ERROR",
    "TRADE_RETCODE_TIMEOUT",
    "TRADE_RETCODE_INVALID",
    "TRADE_RETCODE_INVALID_VOLUME",
    "TRADE_RETCODE_INVALID_PRICE",
    "TRADE_RETCODE_INVALID_STOPS",
    "TRADE_RETCODE_TRADE_DISABLED",
    "TRADE_RETCODE_MARKET_CLOSED",
    "TRADE_RETCODE_NO_MONEY",
    "TRADE_RETCODE_PRICE_CHANGED",
    "TRADE_RETCODE_PRICE_OFF",
    "TRADE_RETCODE_INVALID_EXPIRATION",
    "TRADE_RETCODE_ORDER_CHANGED",
    "TRADE_RETCODE_TOO_MANY_REQUESTS",
    "TRADE_RETCODE_NO_CHANGES",
    "TRADE_RETCODE_SERVER_DISABLES_AT",
    "TRADE_RETCODE_CLIENT_DISABLES_AT",
    "TRADE_RETCODE_LOCKED",
    "TRADE_RETCODE_FROZEN",
    "TRADE_RETCODE_INVALID_FILL",
    "TRADE_RETCODE_CONNECTION",
    "TRADE_RETCODE_ONLY_REAL",
    "TRADE_RETCODE_LIMIT_ORDERS",
    "TRADE_RETCODE_LIMIT_VOLUME",
    "TRADE_RETCODE_INVALID_ORDER",
    "TRADE_RETCODE_POSITION_CLOSED",
    "TRADE_RETCODE_INVALID_CLOSE_VOLUME",
    "TRADE_RETCODE_CLOSE_ORDER_EXIST",
    "TRADE_RETCODE_LIMIT_POSITIONS",
    "TRADE_RETCODE_REJECT_CANCEL",
    "TRADE_RETCODE_LONG_ONLY",
    "TRADE_RETCODE_SHORT_ONLY",
    "TRADE_RETCODE_CLOSE_ONLY",
    "TRADE_RETCODE_FIFO_CLOSE",
    "RES_S_OK",
    "RES_E_FAIL",
    "RES_E_INVALID_PARAMS",
    "RES_E_NO_MEMORY",
    "RES_E_NOT_FOUND",
    "RES_E_INVALID_VERSION",
    "RES_E_AUTH_FAILED",
    "RES_E_UNSUPPORTED",
    "RES_E_AUTO_TRADING_DISABLED",
    "RES_E_INTERNAL_FAIL",
    "RES_E_INTERNAL_FAIL_SEND",
    "RES_E_INTERNAL_FAIL_RECEIVE",
    "RES_E_INTERNAL_FAIL_INIT",
    "RES_E_INTERNAL_FAIL_CONNECT",
    "RES_E_INTERNAL_FAIL_TIMEOUT",
)
core_mt5_functions = (
    "initialize",
    "shutdown",
    "login",
    "version",
    "terminal_info",
    "account_info",
    "copy_ticks_from",
    "copy_ticks_range",
    "copy_rates_from",
    "copy_rates_from_pos",
    "copy_rates_range",
    "positions_total",
    "positions_get",
    "orders_total",
    "orders_get",
    "history_orders_total",
    "history_orders_get",
    "history_deals_total",
    "history_deals_get",
    "order_check",
    "order_send",
    "order_calc_margin",
    "order_calc_profit",
    "symbol_info",
    "symbol_info_tick",
    "symbol_select",
    "symbols_total",
    "symbols_get",
    "market_book_add",
    "market_book_release",
    "market_book_get",
    "last_error",
)
types = (
    "TradePosition",
    "TradeOrder",
    "TradeDeal",
    "TradeRequest",
    "OrderSendResult",
    "OrderCheckResult",
    "Tick",
    "TerminalInfo",
    "SymbolInfo",
    "AccountInfo",
    "BookInfo",
)


class BaseMeta(type):
    def __new__(mcs, cls_name, bases, cls_dict):
        defaults: dict = getattr(MetaTrader5, "__dict__", {})
        callables = {f"_{key}": value for key in core_mt5_functions if (value := defaults.get(key, None)) is not None}
        consts = {key: value for key in constants if (value := defaults.get(key, None)) is not None}
        types_ = {key: value for key in types if (value := defaults.get(key, None)) is not None}
        cls_dict |= callables
        cls_dict |= consts
        cls_dict |= types_
        return super().__new__(mcs, cls_name, bases, cls_dict)


class MetaCore(metaclass=BaseMeta):
    TIMEFRAME_M1: int
    TIMEFRAME_M2: int
    TIMEFRAME_M3: int
    TIMEFRAME_M4: int
    TIMEFRAME_M5: int
    TIMEFRAME_M6: int
    TIMEFRAME_M10: int
    TIMEFRAME_M12: int
    TIMEFRAME_M15: int
    TIMEFRAME_M20: int
    TIMEFRAME_M30: int
    TIMEFRAME_H1: int
    TIMEFRAME_H2: int
    TIMEFRAME_H4: int
    TIMEFRAME_H3: int
    TIMEFRAME_H6: int
    TIMEFRAME_H8: int
    TIMEFRAME_H12: int
    TIMEFRAME_D1: int
    TIMEFRAME_W1: int
    TIMEFRAME_MN1: int
    COPY_TICKS_ALL: int
    COPY_TICKS_INFO: int
    COPY_TICKS_TRADE: int
    TICK_FLAG_BID: int
    TICK_FLAG_ASK: int
    TICK_FLAG_LAST: int
    TICK_FLAG_VOLUME: int
    TICK_FLAG_BUY: int
    TICK_FLAG_SELL: int
    POSITION_TYPE_BUY: int
    POSITION_TYPE_SELL: int
    POSITION_REASON_CLIENT: int
    POSITION_REASON_MOBILE: int
    POSITION_REASON_WEB: int
    POSITION_REASON_EXPERT: int
    ORDER_TYPE_BUY: int
    ORDER_TYPE_SELL: int
    ORDER_TYPE_BUY_LIMIT: int
    ORDER_TYPE_SELL_LIMIT: int
    ORDER_TYPE_BUY_STOP: int
    ORDER_TYPE_SELL_STOP: int
    ORDER_TYPE_BUY_STOP_LIMIT: int
    ORDER_TYPE_SELL_STOP_LIMIT: int
    ORDER_TYPE_CLOSE_BY: int
    ORDER_STATE_STARTED: int
    ORDER_STATE_PLACED: int
    ORDER_STATE_CANCELED: int
    ORDER_STATE_PARTIAL: int
    ORDER_STATE_FILLED: int
    ORDER_STATE_REJECTED: int
    ORDER_STATE_EXPIRED: int
    ORDER_STATE_REQUEST_ADD: int
    ORDER_STATE_REQUEST_MODIFY: int
    ORDER_STATE_REQUEST_CANCEL: int
    ORDER_FILLING_FOK: int
    ORDER_FILLING_IOC: int
    ORDER_FILLING_RETURN: int
    ORDER_FILLING_BOC: int
    ORDER_TIME_GTC: int
    ORDER_TIME_DAY: int
    ORDER_TIME_SPECIFIED: int
    ORDER_TIME_SPECIFIED_DAY: int
    ORDER_REASON_CLIENT: int
    ORDER_REASON_MOBILE: int
    ORDER_REASON_WEB: int
    ORDER_REASON_EXPERT: int
    ORDER_REASON_SL: int
    ORDER_REASON_TP: int
    ORDER_REASON_SO: int
    DEAL_TYPE_BUY: int
    DEAL_TYPE_SELL: int
    DEAL_TYPE_BALANCE: int
    DEAL_TYPE_CREDIT: int
    DEAL_TYPE_CHARGE: int
    DEAL_TYPE_CORRECTION: int
    DEAL_TYPE_BONUS: int
    DEAL_TYPE_COMMISSION: int
    DEAL_TYPE_COMMISSION_DAILY: int
    DEAL_TYPE_COMMISSION_MONTHLY: int
    DEAL_TYPE_COMMISSION_AGENT_DAILY: int
    DEAL_TYPE_COMMISSION_AGENT_MONTHLY: int
    DEAL_TYPE_INTEREST: int
    DEAL_TYPE_BUY_CANCELED: int
    DEAL_TYPE_SELL_CANCELED: int
    DEAL_DIVIDEND: int
    DEAL_DIVIDEND_FRANKED: int
    DEAL_TAX: int
    DEAL_ENTRY_IN: int
    DEAL_ENTRY_OUT: int
    DEAL_ENTRY_INOUT: int
    DEAL_ENTRY_OUT_BY: int
    DEAL_REASON_CLIENT: int
    DEAL_REASON_MOBILE: int
    DEAL_REASON_WEB: int
    DEAL_REASON_EXPERT: int
    DEAL_REASON_SL: int
    DEAL_REASON_TP: int
    DEAL_REASON_SO: int
    DEAL_REASON_ROLLOVER: int
    DEAL_REASON_VMARGIN: int
    DEAL_REASON_SPLIT: int
    TRADE_ACTION_DEAL: int
    TRADE_ACTION_PENDING: int
    TRADE_ACTION_SLTP: int
    TRADE_ACTION_MODIFY: int
    TRADE_ACTION_REMOVE: int
    TRADE_ACTION_CLOSE_BY: int
    SYMBOL_CHART_MODE_BID: int
    SYMBOL_CHART_MODE_LAST: int
    SYMBOL_CALC_MODE_FOREX: int
    SYMBOL_CALC_MODE_FUTURES: int
    SYMBOL_CALC_MODE_CFD: int
    SYMBOL_CALC_MODE_CFDINDEX: int
    SYMBOL_CALC_MODE_CFDLEVERAGE: int
    SYMBOL_CALC_MODE_FOREX_NO_LEVERAGE: int
    SYMBOL_CALC_MODE_EXCH_STOCKS: int
    SYMBOL_CALC_MODE_EXCH_FUTURES: int
    SYMBOL_CALC_MODE_EXCH_OPTIONS: int
    SYMBOL_CALC_MODE_EXCH_OPTIONS_MARGIN: int
    SYMBOL_CALC_MODE_EXCH_BONDS: int
    SYMBOL_CALC_MODE_EXCH_STOCKS_MOEX: int
    SYMBOL_CALC_MODE_EXCH_BONDS_MOEX: int
    SYMBOL_CALC_MODE_SERV_COLLATERAL: int
    SYMBOL_TRADE_MODE_DISABLED: int
    SYMBOL_TRADE_MODE_LONGONLY: int
    SYMBOL_TRADE_MODE_SHORTONLY: int
    SYMBOL_TRADE_MODE_CLOSEONLY: int
    SYMBOL_TRADE_MODE_FULL: int
    SYMBOL_TRADE_EXECUTION_REQUEST: int
    SYMBOL_TRADE_EXECUTION_INSTANT: int
    SYMBOL_TRADE_EXECUTION_MARKET: int
    SYMBOL_TRADE_EXECUTION_EXCHANGE: int
    SYMBOL_SWAP_MODE_DISABLED: int
    SYMBOL_SWAP_MODE_POINTS: int
    SYMBOL_SWAP_MODE_CURRENCY_SYMBOL: int
    SYMBOL_SWAP_MODE_CURRENCY_MARGIN: int
    SYMBOL_SWAP_MODE_CURRENCY_DEPOSIT: int
    SYMBOL_SWAP_MODE_INTEREST_CURRENT: int
    SYMBOL_SWAP_MODE_INTEREST_OPEN: int
    SYMBOL_SWAP_MODE_REOPEN_CURRENT: int
    SYMBOL_SWAP_MODE_REOPEN_BID: int
    DAY_OF_WEEK_SUNDAY: int
    DAY_OF_WEEK_MONDAY: int
    DAY_OF_WEEK_TUESDAY: int
    DAY_OF_WEEK_WEDNESDAY: int
    DAY_OF_WEEK_THURSDAY: int
    DAY_OF_WEEK_FRIDAY: int
    DAY_OF_WEEK_SATURDAY: int
    SYMBOL_ORDERS_GTC: int
    SYMBOL_ORDERS_DAILY: int
    SYMBOL_ORDERS_DAILY_NO_STOPS: int
    SYMBOL_OPTION_RIGHT_CALL: int
    SYMBOL_OPTION_RIGHT_PUT: int
    SYMBOL_OPTION_MODE_EUROPEAN: int
    SYMBOL_OPTION_MODE_AMERICAN: int
    ACCOUNT_TRADE_MODE_DEMO: int
    ACCOUNT_TRADE_MODE_CONTEST: int
    ACCOUNT_TRADE_MODE_REAL: int
    ACCOUNT_STOPOUT_MODE_PERCENT: int
    ACCOUNT_STOPOUT_MODE_MONEY: int
    ACCOUNT_MARGIN_MODE_RETAIL_NETTING: int
    ACCOUNT_MARGIN_MODE_EXCHANGE: int
    ACCOUNT_MARGIN_MODE_RETAIL_HEDGING: int
    BOOK_TYPE_SELL: int
    BOOK_TYPE_BUY: int
    BOOK_TYPE_SELL_MARKET: int
    BOOK_TYPE_BUY_MARKET: int
    TRADE_RETCODE_REQUOTE: int
    TRADE_RETCODE_REJECT: int
    TRADE_RETCODE_CANCEL: int
    TRADE_RETCODE_PLACED: int
    TRADE_RETCODE_DONE: int
    TRADE_RETCODE_DONE_PARTIAL: int
    TRADE_RETCODE_ERROR: int
    TRADE_RETCODE_TIMEOUT: int
    TRADE_RETCODE_INVALID: int
    TRADE_RETCODE_INVALID_VOLUME: int
    TRADE_RETCODE_INVALID_PRICE: int
    TRADE_RETCODE_INVALID_STOPS: int
    TRADE_RETCODE_TRADE_DISABLED: int
    TRADE_RETCODE_MARKET_CLOSED: int
    TRADE_RETCODE_NO_MONEY: int
    TRADE_RETCODE_PRICE_CHANGED: int
    TRADE_RETCODE_PRICE_OFF: int
    TRADE_RETCODE_INVALID_EXPIRATION: int
    TRADE_RETCODE_ORDER_CHANGED: int
    TRADE_RETCODE_TOO_MANY_REQUESTS: int
    TRADE_RETCODE_NO_CHANGES: int
    TRADE_RETCODE_SERVER_DISABLES_AT: int
    TRADE_RETCODE_CLIENT_DISABLES_AT: int
    TRADE_RETCODE_LOCKED: int
    TRADE_RETCODE_FROZEN: int
    TRADE_RETCODE_INVALID_FILL: int
    TRADE_RETCODE_CONNECTION: int
    TRADE_RETCODE_ONLY_REAL: int
    TRADE_RETCODE_LIMIT_ORDERS: int
    TRADE_RETCODE_LIMIT_VOLUME: int
    TRADE_RETCODE_INVALID_ORDER: int
    TRADE_RETCODE_POSITION_CLOSED: int
    TRADE_RETCODE_INVALID_CLOSE_VOLUME: int
    TRADE_RETCODE_CLOSE_ORDER_EXIST: int
    TRADE_RETCODE_LIMIT_POSITIONS: int
    TRADE_RETCODE_REJECT_CANCEL: int
    TRADE_RETCODE_LONG_ONLY: int
    TRADE_RETCODE_SHORT_ONLY: int
    TRADE_RETCODE_CLOSE_ONLY: int
    TRADE_RETCODE_FIFO_CLOSE: int
    RES_S_OK: int
    RES_E_FAIL: int
    RES_E_INVALID_PARAMS: int
    RES_E_NO_MEMORY: int
    RES_E_NOT_FOUND: int
    RES_E_INVALID_VERSION: int
    RES_E_AUTH_FAILED: int
    RES_E_UNSUPPORTED: int
    RES_E_AUTO_TRADING_DISABLED: int
    RES_E_INTERNAL_FAIL: int
    RES_E_INTERNAL_FAIL_SEND: int
    RES_E_INTERNAL_FAIL_RECEIVE: int
    RES_E_INTERNAL_FAIL_INIT: int
    RES_E_INTERNAL_FAIL_CONNECT: int
    RES_E_INTERNAL_FAIL_TIMEOUT: int
    _account_info: Callable
    _copy_rates_from: Callable
    _copy_rates_from_pos: Callable
    _copy_rates_range: Callable
    _copy_ticks_from: Callable
    _copy_ticks_range: Callable
    _history_deals_get: Callable
    _history_deals_total: Callable
    _history_orders_get: Callable
    _history_orders_total: Callable
    _initialize: Callable
    _last_error: Callable
    _login: Callable
    _market_book_add: Callable
    _market_book_get: Callable
    _market_book_release: Callable
    _order_calc_margin: Callable
    _order_calc_profit: Callable
    _order_check: Callable
    _order_send: Callable
    _orders_get: Callable
    _orders_total: Callable
    _positions_get: Callable
    _positions_total: Callable
    _shutdown: Callable
    _symbol_info: Callable
    _symbol_info_tick: Callable
    _symbol_select: Callable
    _symbols_get: Callable
    _symbols_total: Callable
    _terminal_info: Callable
    _version: Callable
    config: Config
    AccountInfo: AccountInfo
    TradePosition: TradePosition
    TradeOrder: TradeOrder
    TradeDeal: TradeDeal
    TradeRequest: TradeRequest
    OrderSendResult: OrderSendResult
    OrderCheckResult: OrderCheckResult
    Tick: Tick
    TerminalInfo: TerminalInfo
    SymbolInfo: SymbolInfo
    BookInfo: BookInfo



from logging import getLogger

from ...core.models import OrderType
from ...lib.trader import Trader

logger = getLogger(__name__)


class ScalpTrader(Trader):
    async def place_trade(self, *, order_type: OrderType, volume: float = None, parameters: dict = None):
        """Places a trade based on the order_type and volume. The volume is optional. If not provided, the minimum volume
        for the symbol will be used. This trade is placed without a stop_loss or take_profit. The trade is recorded in the
        trade_record file.

        Args:
            order_type (OrderType): The order_type
            volume (float): The volume to trade
            parameters (dict): Parameters associated with the trade
        """
        try:
            self.parameters |= parameters or {}
            volume = volume or self.symbol.volume_min
            await self.create_order_no_stops(order_type=order_type, volume=volume)
            if not await self.check_order():
                return
            self.order.comment = self.parameters.get("name", self.__class__.__name__)
            res = await self.send_order()
            if res is not None:
                await self.record_trade(result=res, parameters=self.parameters)
        except Exception as err:
            logger.error(f"{err} in {self.__class__.__name__}.place_trade for {self.symbol.name}")



class Error:
    """Error class for handling errors"""

    descriptions = {
        # common errors
        1: "Successful",
        -1: "generic fail",
        -2: "invalid arguments/parameters",
        -3: "no memory condition",
        -4: "no history",
        -5: "invalid version",
        -6: "authorization failed",
        -7: "unsupported method",
        -8: "auto-trading disabled",
        # internal errors
        -10000: "internal IPC general error",
        -10001: "internal IPC send failed",
        -10002: "internal IPC recv failed",
        -10003: "internal IPC initialization fail",
        -10004: "internal IPC no ipc",
        -10005: "internal timeout",
    }

    conn_errors = (-10000, -10001, -10002, -10003, -10004, -10005, -6)

    def __init__(self, code: int, description: str = ""):
        self.code = code
        self.description = description or self.descriptions.get(code, "unknown error")

    def is_connection_error(self):
        return self.code in self.conn_errors

    def __repr__(self):
        return f"{self.code}: {self.description}"



from .account import Account
from .bot import Bot
from .candle import Candle, Candles
from .executor import Executor
from .history import History
from .order import Order
from .positions import Positions
from .ram import RAM
from .result import Result
from .symbol import Symbol
from .ticks import Tick, Ticks
from .trader import Trader
from .strategy import Strategy
from .sessions import Sessions, Session
from .trade_records import TradeRecords
from .terminal import Terminal
from .backtester import BackTester
